#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <netdb.h>

#include <ucd-snmp-config.h>
#include <asn1.h>
#include <snmp.h>
#include <snmp_impl.h>
#include <snmp_api.h>
#include <snmp_client.h>
#include <mib.h>

#include "quickfind.h"
#include "fsnmp.h"
#include "list.h"

#define DEFAULT_PACKETS_MAX_SENT	9
#define DEFAULT_COMMUNITY		"public"
#define DEFAULT_SNMP_VERSION		SNMP_VERSION_1
#define DEFAULT_SNMP_PORT		161
#define DEFAULT_TIMEOUT			1000
#define DEFAULT_BACKOFF			0
#define DEFAULT_MAXRETRANSMITS		5

#define MAX_PROCESS_TIMEOUT		5
#define MAX_SEQUENTIAL_PACKETS		20
#define MAX_SEQUENTIAL_LINES		50

#define INBUF_SIZE			16384
#define PACKET_SIZE 			4096

#undef DEBUG

#ifdef DEBUG
#define DBGPRINT( a... ) 	fprintf(stderr, a)
#else
#define DBGPRINT( a... )
#endif

int fsnmp_send_packet(struct fsnmp_packet *fsp);
void fsnmp_close_session(struct fsnmp_session *s);

/*
 *
 * GLOBALS
 *
 */
static char			*inbuf;
static struct fsnmp_transport	*transport;
LIST_HEAD(struct fsnmp_packet)	event_queue;
LIST_HEAD(struct fsnmp_session)	session_list;



#ifdef DEBUG
void fsnmp_dump_sessions(void ) {
	struct fsnmp_session *s;

	LIST_WALK(&session_list, s) {
		printf("%-20s %d %d %6s %-10s\n", 
			s->host, s->ppending, s->ptotal,
			(LIST_EMPTY(&s->packets_queued) ? "empty" : "full"),
			s->finished ? "finished" : "unfinished");
	}
}
#endif

int dump_packet(struct fsnmp_packet *p, struct snmp_pdu *pdu, struct timeval *tv) {
	struct variable_list    *vp;
	char			buf[256];
			
	/* SNMP error (nosuchname etc) */
	if (pdu->errstat != SNMP_ERR_NOERROR) {
		fprintf(stdout, "%lu.%03lu %s error %s\n",
			tv->tv_sec, tv->tv_usec / 1000,
			p->ureqid, snmp_errstring(pdu->errstat));
		return(0);
	} 

	/* Print our variables */
	for(vp=pdu->variables;vp;vp=vp->next_variable) {
		/* If we got an oid shorter than we requested
		 * or we got a completely bogus answer */
		if (vp->name_length < p->oidlen ||
			(memcmp(p->oid, vp->name, p->oidlen * sizeof(oid)) != 0)) {

			if (p->pdu->command == SNMP_MSG_GET) {
				fprintf(stdout, "%lu.%03lu %s error got bogus result var\n",
						tv->tv_sec, tv->tv_usec / 1000, p->ureqid);
			} else {
				 fprintf(stdout, "%lu.%03lu %s success EOT\n",
						tv->tv_sec, tv->tv_usec / 1000, p->ureqid);
			}

			return(0);
		} else {
			/* Decode value to name */
			sprint_variable(buf, vp->name, vp->name_length, vp);
			fprintf(stdout, "%lu.%03lu %s success %s\n",
				tv->tv_sec, tv->tv_usec / 1000, p->ureqid, buf);

		}
	}

	return(1);
}

int fsnmp_parse_version(u_char *packet, size_t length) {
	u_char		type;
	long 		version;

	packet = asn_parse_sequence(packet, &length, &type,
			(ASN_SEQUENCE | ASN_CONSTRUCTOR), "version");

	if (packet) {
		packet = asn_parse_int(packet, &length, &type, &version, sizeof(version));
		if (!packet || type != ASN_INTEGER) {
			return(-1);
		}
	}

	return(version);
}

void fsnmp_free_packet(struct fsnmp_packet *p) {

	if (p->ureqid)
		free(p->ureqid);
	if (p->pdu)
		snmp_free_pdu(p->pdu);

#ifdef DEBUG
	/* Poison session pointer in free packet */
	p->session=(void *) 0xdeadbeef;
#endif

	free(p);
}

void fsnmp_process_response(int action, struct snmp_pdu *pdu, struct fsnmp_packet *p) {
	struct fsnmp_session 	*s;
	struct variable_list	*vp, *pl;	
	struct timeval		tv;

	gettimeofday(&tv, NULL);

	s=p->session;
	s->ppending--;

	switch(action) {
		case(RECEIVED_MESSAGE):

			dump_packet(p, pdu, &tv);

			switch(p->pdu->command) {
				case(SNMP_MSG_GETNEXT):
				
					/* Check for getnext vars */
					pl=p->pdu->variables;
					p->pdu->variables=NULL;

					/* Print our variables */
					for(vp=pdu->variables;vp;vp=vp->next_variable) {
						/* If we got an oid shorter than we requested
						 * or we got a completely bogus answer */
						if (vp->name_length >= p->oidlen &&
							(memcmp(p->oid, vp->name, p->oidlen * sizeof(oid)) == 0)) {

							/* Add request to next MSG_GET_NEXT packet */
							snmp_add_null_var(p->pdu, vp->name, vp->name_length);
						}
					}

					/* Free old variabled */
					snmp_free_varbind(pl); 

					/* Did we find variables to do a GETNEXT on ? */
					if (!p->pdu->variables)
						break;

					/* Get new ids */
					p->pdu->msgid=snmp_get_next_msgid();
					p->pdu->reqid=snmp_get_next_reqid();

					p->retransmits=0;

					fsnmp_send_packet(p);

					/* We return as we dont want this packet to be deleted */
					return;

				case(SNMP_MSG_GET):
					/* We dumped the packet already - Just fall through */
					break;

				default:
					fprintf(stderr, "%s: Interesting packet - unknown pdu->command %x in request\n",
							__FUNCTION__, p->pdu->command);
					break;
			}
			break;
		case(TIMED_OUT):
			fprintf(stdout, "%lu.%03lu %s error timed out\n", 
					tv.tv_sec, tv.tv_usec / 1000, p->ureqid);
			break;
		default:
			fprintf(stdout, "%lu.%03lu %s error unknown error %d\n", 
					tv.tv_sec, tv.tv_usec / 1000, p->ureqid, action);
			break;
	}

	s->ptotal--;

	/* Remove from event or queued list */
	LIST_DELETE(p);
	fsnmp_free_packet(p);

	/* Try sending queued packets */
	while(!LIST_EMPTY(&s->packets_queued) &&
			s->ppending < s->ppendingmax) {

		if (!s->packets_queued.next)
			fprintf(stderr, "%s: list not empty but first element is NULL\n", __FUNCTION__);
		if (fsnmp_send_packet(s->packets_queued.next))
			break;
	}

	/* Close session if no work left and closed in input */
	if(!s->ptotal && s->finished) 
		fsnmp_close_session(s);
}

void fsnmp_receive_packet(struct fsnmp_transport *t) {
	static u_char		packet[PACKET_SIZE];
	static u_char		community[COMMUNITY_MAX_LEN];
	u_char			*pptr;
	struct sockaddr		from;
	struct fsnmp_packet	*p;
	struct snmp_pdu		pdu;
	int			comlen,
				plen,
				flen,
				i;

	/* We dont want to spent all our time in this loop
	 * so be carefull to exit one day 		*/
	for(i=0;i<MAX_SEQUENTIAL_PACKETS;i++) {
		flen=sizeof(from);

		plen=recvfrom(t->socket, packet, PACKET_SIZE, 0, &from, &flen);

		if (plen<=0)
			break;

		memset(&pdu, 0, sizeof(struct snmp_pdu));

		pptr=packet;

		pdu.version=fsnmp_parse_version(pptr, plen);

		comlen=sizeof(community);
		pptr = snmp_comstr_parse(pptr, &plen, community,
				&comlen, &pdu.version);
		assert((pdu.community=strdup(community)));

		flen=snmp_pdu_parse(&pdu, pptr, &plen);

		/* Get fsnmp_packet from lookup hash by reqid */
		p=deletekey(t->ppreqid, pdu.reqid);

		if(p)
			fsnmp_process_response(RECEIVED_MESSAGE, &pdu, p);
#ifdef DEBUG
		if(!p) 
			fprintf(stderr, "%s: Oops - unknown packet reqid %lx msgid %lx\n",
					__FUNCTION__, pdu.reqid, pdu.msgid);
#endif
		free(pdu.community);
	}

	return;
}

void fsnmp_process_fdset(fd_set *fdset) {
	if (FD_ISSET(transport->socket, fdset)) {
		fsnmp_receive_packet(transport);
	}
}

struct fsnmp_transport *fsnmp_get_transport(void ) {
	return(transport);
}

int fsnmp_open_transport(struct fsnmp_transport *t) {
	int	fd;
	struct sockaddr_in	saddr;

	memset(&saddr, 0, sizeof(saddr));

	assert((fd=socket(AF_INET, SOCK_DGRAM, 0)));
	assert((bind(fd, (struct sockaddr *) &saddr, sizeof(saddr)) == 0));

	t->socket=fd;

	t->ppreqid=createqf();

	/* Set socket non-blocking */
	fcntl(t->socket, F_SETFL, O_NONBLOCK);

	return(0);
}

void fsnmp_close_session(struct fsnmp_session *s) {

	LIST_DELETE(s);

	if (s->host)
		free(s->host);

	if (s->community)
		free(s->community);

	if (!LIST_EMPTY(&s->packets_queued))
		fprintf(stderr, "%s: closing session although packet queue not empty\n",
				__FUNCTION__);

	if (!s->finished) {
		fprintf(stderr, "%s: closing session although not finished\n",
				__FUNCTION__);
		assert(s->finished);
	}

	if (s->ppending)
		fprintf(stderr, "%s: closing session although %d packets pending\n",
				__FUNCTION__, s->ppending);

	if (s->ptotal) 
		fprintf(stderr, "%s: closing session although %d packets total\n",
				__FUNCTION__, s->ptotal);

	s->transport->refcount--;

#ifdef DEBUG
	/* Poison transport pointer to see closed sessions in gdb */
	s->transport=(void *) 0xdeadbeef;
#endif

	free(s);
}

void fsnmp_close_transport(struct fsnmp_transport *t) {
	close(t->socket);
}

int fsnmp_send_packet(struct fsnmp_packet *fsp) {
	static u_char	packet[PACKET_SIZE],
			*ps;
	int		plen=PACKET_SIZE,
			eno,
			bs;
	unsigned long	tom;

	if (!fsp) {
		fprintf(stderr, "%s: got NULL packet passed\n", __FUNCTION__);
		return(0);
	}

	ps=fsnmp_build_packet(fsp->pdu, packet, 
				&plen, &eno);
	if (!ps) {
		fprintf(stderr, "%s: error: %s\n", __FUNCTION__, snmp_errstring(eno));
		return(-1);
	}

	bs=sendto(fsp->session->transport->socket, ps, plen, 0, 
			(struct sockaddr *) &fsp->session->saddr,
			sizeof(fsp->session->saddr));

	if (bs!=plen) {
		fprintf(stderr, "%s: Send failed for %s/%lx\n", __FUNCTION__,
				fsp->session->host, fsp->pdu->reqid);
		return(-1);
	}

	fsp->session->ppending++;

	/* Store sent time */
	gettimeofday(&fsp->sent, NULL);

	/* Calc timeout in ms */
	tom=fsp->session->timeout+
		fsp->retransmits*fsp->session->backoff;

	fsp->sent.tv_sec+=(tom / 1000);
	fsp->sent.tv_usec+=(tom % 1000) * 1000;

	fsp->retransmits++;

	insertkey(fsp->session->transport->ppreqid,
			fsp->pdu->reqid, (void *) fsp);

	/* Remove from queued or event list */
	LIST_DELETE(fsp);

	/* Append to event queue */
	LIST_APPEND(&event_queue, fsp);

	return(0);
}


#define PL_MAX_ARGS	20
struct parsed_line {
	int	argc;	
	char	*argv[PL_MAX_ARGS];
};

struct parsed_line *parse_line(char *line) {
	static struct parsed_line	pl;
	int				i,j;

	pl.argc=0;

	j=strlen(line);
	for(i=0;i<j;i++) {
		if(line[i] == '"') {
			for(i++;i<j && line[i] != '"';i++);
			i++;
		}

		for(;i<j && line[i] == ' ';line[i++]=0x0);

		pl.argv[pl.argc++]=&line[i];

		for(;i<j && line[i] != ' '
			&& line[i] != '"';i++);

		i--;
	}

	return(&pl);
}


/*
 * session host=127.0.0.1 community=public version=2 timeout=5
 * id get|walk oid
 * end
 */
struct fsnmp_session *create_session(struct parsed_line *pl) {
	int 			i;
	char			*eq;
	struct fsnmp_session	*s;
	struct hostent		*he;

	DBGPRINT("%s:%d\n", __FUNCTION__, __LINE__);

	assert((s=(void *) malloc(sizeof(struct fsnmp_session))));
	memset(s, 0, sizeof(struct fsnmp_session));

	s->ppendingmax=DEFAULT_PACKETS_MAX_SENT;
	s->version=DEFAULT_SNMP_VERSION;
	s->port=DEFAULT_SNMP_PORT;
	s->timeout=DEFAULT_TIMEOUT;
	s->backoff=DEFAULT_BACKOFF;
	s->maxretransmits=DEFAULT_MAXRETRANSMITS;

#define SESSION_PARM_HOST	"host"
#define SESSION_PARM_COMMUNITY	"community"
#define SESSION_PARM_VERSION	"version"

	for(i=1;i<pl->argc;i++) {
		if (strncmp(pl->argv[i], SESSION_PARM_HOST, strlen(SESSION_PARM_HOST)) == 0) {
			if ((eq=strstr(pl->argv[i], "="))) {
				s->host=strdup(&eq[1]);
			}
		} else if (strncmp(pl->argv[i], SESSION_PARM_COMMUNITY,
					strlen(SESSION_PARM_COMMUNITY)) == 0) {
			if ((eq=strstr(pl->argv[i], "="))) {
				s->community=strdup(&eq[1]);
			}
		} else if (strncmp(pl->argv[i], SESSION_PARM_VERSION,
					strlen(SESSION_PARM_VERSION)) == 0) {
			if ((eq=strstr(pl->argv[i], "="))) {

				/* Skip space */
				eq++;

				if (strcmp(eq,"1") == 0)
					s->version=SNMP_VERSION_1;
				else if (strcmp(eq,"2") == 0)
					s->version=SNMP_VERSION_2c;
				else
					fprintf(stderr, "Unknown snmp version %s\n", eq);
			}
		} else {
			fprintf(stderr, "Syntax error\n");
		}
	}

	if (!s->host) {
		fprintf(stderr, "No hostname/ip address\n");
		if (s->community)
			free(s->community);
		free(s);
		return(NULL);
	}

	if (!s->community) 
		s->community=strdup(DEFAULT_COMMUNITY);

	s->transport=fsnmp_get_transport();

	he=gethostbyname(s->host);

	if (!he) {
		free(s->community);
		free(s->host);
		free(s);
		return(NULL);
	}

	memcpy(&s->saddr.sin_addr, he->h_addr_list[0], he->h_length);
	s->saddr.sin_port=htons(s->port);

	s->transport->refcount++;

	LIST_INIT(&s->packets_queued);

	LIST_ADD(&session_list, s);

	return(s);
}

int add_request(struct fsnmp_session *s, struct parsed_line *pl) {
	struct fsnmp_packet	*p;
	int			reqtype;


#define REQTYPE_GET	"get"
#define REQTYPE_WALK	"walk"

	if (strcmp(pl->argv[1], REQTYPE_GET) == 0) {
		reqtype=SNMP_MSG_GET;
	} else if (strcmp(pl->argv[1], REQTYPE_WALK) == 0) {
		reqtype=SNMP_MSG_GETNEXT;
	} else {
		fprintf(stderr, "Unknown request type %s\n", pl->argv[1]);
		return(1);
	}


	assert((p=(void *)malloc(sizeof(struct fsnmp_packet))));
	memset(p, 0, sizeof(struct fsnmp_packet));

	p->session=s;

	p->pdu=snmp_pdu_create(reqtype);

	p->pdu->community=strdup(p->session->community);
	p->pdu->community_len=strlen(p->pdu->community);
	p->pdu->version=p->session->version;

	p->ureqid=strdup(pl->argv[0]);

	p->oidlen=MAX_OID_LEN;
	if (snmp_parse_oid(pl->argv[2], p->oid, &p->oidlen) == NULL) {
		snmp_perror(pl->argv[2]);

		free(p->ureqid);
		free(p->pdu->community);
		free(p);

		return(1);
	}

	snmp_add_null_var(p->pdu, p->oid, p->oidlen);

	LIST_ADD(&p->session->packets_queued, p);
	p->session->ptotal++;

	if (p->session->ppending >= p->session->ppendingmax) 
		return(0);

	return fsnmp_send_packet(p);
}

void process_line(char *line) {
	static struct fsnmp_session	*s;
	struct parsed_line		*pl;

	pl=parse_line(line);

	if (pl->argc >= 1) {

		if (strcmp(pl->argv[0], "session") == 0) {
			if (s) {
				fprintf(stderr, "Session was not properly ended\n");
			}
			s=create_session(pl);
			return;
		} else if (strcmp(pl->argv[0], "end") == 0) {
			s->finished=1;
			if (!s->ptotal)
				fsnmp_close_session(s);
			s=NULL;
			return;
		}

		if (!s) {
			fprintf(stderr, "Not in session context\n");
			return;
		}

		if (pl->argc != 3) {
			fprintf(stderr, "Wrong number of args in session context\n");
			return;
		}

		add_request(s, pl);
	}
}

/* We get called if fd is ready to read */
int readfile() {
	ssize_t		rbytes,
			toread;
	static int	validin;
	int		eof=0,
			sol,
			i,j;

	toread=INBUF_SIZE - validin;

	rbytes=read(fileno(stdin), &inbuf[validin], toread);

	/* If we have an error or read nothing we are boned aka EOF */
	if ((rbytes == -1 && errno != 0) || rbytes < toread)
		eof=1;

	/* update valid bytes in inbuffer */
	validin+=rbytes;

	sol=0;
	/* We dont want to hog the CPU here - So be shure to
	 * exit one day */
	for(j=0;j<MAX_SEQUENTIAL_LINES;j++) {
		for(i=sol;i<=validin && inbuf[i] != 0xa;i++);

		if (inbuf[i]==0xa || (eof && i==validin)) {
			inbuf[i]=0x0;

			process_line(&inbuf[sol]);

			sol=i+1;
		} else 
			break;
	}

	/* Still input to process ? */
	if (sol >= validin) {
		validin=0;
	} else {

		/* Calculate leftover bytes in buffer */
		i=validin-sol;

		/* Move half line in buffer to front */
		memmove(inbuf, &inbuf[sol], i);
		validin=i;
	}

	return(eof);
}


int isfifo(int fd) {
	struct stat	st;

	fstat(fd, &st);

	return S_ISFIFO(st.st_mode);
}

void fsnmp_get_transport_fdset(fd_set *fdset, int *fds) {
	struct fsnmp_transport	*t;

	/* Currently we only have one transport */
	t=fsnmp_get_transport();

	FD_SET(t->socket, fdset);

	if (*fds<=t->socket)
		*fds=t->socket+1;
}


void fsnmp_process_timeout() {
	struct fsnmp_packet	*p, *pp;
	struct timeval		n;
	int			i=0;

	gettimeofday(&n, NULL);

	LIST_WALK(&event_queue, p) {
		pp=p->prev;

		DBGPRINT("%s: now: %lu.%lu tout: %lu.%lu ret: %d reqid: %lx\n", 
				__FUNCTION__, n.tv_sec, n.tv_usec, p->sent.tv_sec, p->sent.tv_usec,
				p->retransmits, p->pdu->reqid);

		if (n.tv_sec > p->sent.tv_sec ||
			(n.tv_sec == p->sent.tv_sec &&
			 n.tv_usec > p->sent.tv_usec)) {
		
			/* Delete key from packet per request id 
			 * (quick lookup table for transport */
			deletekey(p->session->transport->ppreqid, p->pdu->reqid);

			if (p->retransmits >= p->session->maxretransmits) {
				/* This one deletes the packet from all lists
				 * so we need to fake the for counter to not 
				 * segfault dereferencing it the next time */
				fsnmp_process_response(TIMED_OUT, NULL, p);

				p=pp;
				i=0;
			} else {
				p->session->ppending--;

				DBGPRINT("%s: timed out %lx\n", 
						__FUNCTION__, p->pdu->reqid);

				/* Increase the msgid for responses */
				p->pdu->msgid=snmp_get_next_msgid();

				/* Send packet */
				fsnmp_send_packet(p);
				p=pp;
				i=0;
			}
		}

		/* We dont want to traverse the whole list - Just the timeouts
		 * at the beginning and a couple more as we might get "unordered"
		 * in case of packet without backoff beeing queued after packets
		 * with backoff */
		i++;
		if (i>MAX_PROCESS_TIMEOUT)
			break;
	}
}

void process_input() {
	int			fds=0,
				eof=0,
				block,
				stdinflags;
	fd_set			fdset;	
	struct timeval		timeout;


	/* Get stdin's flags */
	stdinflags = fcntl(fileno(stdin), F_GETFL, 0);

	/* FIXME This is strange but works 
	 *
	 * Setting O_NONBLOCK on a fifo results in only 8192 bytes beeing returned
	 *
	 */
	if (!isfifo(fileno(stdin))) {
		DBGPRINT("Setting stdin(%d) to O_NONBLOCK\n", fileno(stdin));
		/* Set stdin non-blocking */
		fcntl(fileno(stdin), F_SETFL, stdinflags | O_NONBLOCK); 
	}

	/* Run until we have an EOF or if we still have sessions */
	while(!eof || !LIST_EMPTY(&session_list)) {

		FD_ZERO(&fdset);
		block=1;

		fsnmp_get_transport_fdset(&fdset, &fds);

		/* We dont need to try reading input when we reached EOF already */
		if (!eof) {
			/* FIXME We could only try stdin if we are low on processing */
			FD_SET(fileno(stdin), &fdset);

			/* If we are stdin only we need to set the maxfd accordingly */
			if (fds < fileno(stdin)+1)
				fds=fileno(stdin)+1;
		}

		/* FIXME: Hardcoded timeout - I dont care polling the
		 * timeout/event table every 200ms but probably someone 
		 * has an idea on how to fix this without walking large
		 * chunks of the event_queue
		 */
		timeout.tv_sec=0;
		timeout.tv_usec=200*1000;

		/* Wait for the sessions and input */
		if ((fds=select(fds, &fdset, NULL, NULL, &timeout))) {

			DBGPRINT("Select returned %d pending fds\n", fds);
	
			/* Do we have input ? */
			if (FD_ISSET(fileno(stdin), &fdset)) {
				DBGPRINT("select meant stdin to be readable\n");
				eof=readfile();
			}

			fsnmp_process_fdset(&fdset);
		}

		/* Process timeout/retransmits in the event_queue */
		fsnmp_process_timeout();
	}

	/* Restore flags in stdin */
	fcntl(fileno(stdin), F_SETFL, stdinflags);
}


int main(int argc, char **argv) {

	init_snmp("fsnmp");

	LIST_INIT(&event_queue);
	LIST_INIT(&session_list);

	assert((transport=malloc(sizeof(struct fsnmp_transport))));	
	fsnmp_open_transport(transport);

	/* Allocate buffer + trailing zero */
	assert((inbuf=(char *) malloc(INBUF_SIZE+1)));

	process_input();

	free(inbuf);
	fsnmp_close_transport(transport);
	free(transport);

	return(0);
}
